@@ -34,9 +34,9 @@ void _IO_ldtostr (long double *, char *, int, int, char);
 #define NBITS ((NI-4)*16)
 
  /* Maximum number of decimal digits in ASCII conversion
-  * = NBITS*log10(2)
+  * Take full possible size of output into account
   */
-#define NDEC (NBITS*8/27)
+#define NDEC 1023
 
  /* The exponent of 1.0 */
 #define EXONE (0x3fff)
@@ -61,13 +61,13 @@ typedef struct
   unsigned short equot[NI];
 } LDPARMS;
 
-static void esub (_CONST short unsigned int *a, _CONST short unsigned int *b,
+static void esub (const short unsigned int *a, const short unsigned int *b,
 		  short unsigned int *c, LDPARMS * ldp);
-static void emul (_CONST short unsigned int *a, _CONST short unsigned int *b,
+static void emul (const short unsigned int *a, const short unsigned int *b,
 		  short unsigned int *c, LDPARMS * ldp);
-static void ediv (_CONST short unsigned int *a, _CONST short unsigned int *b,
+static void ediv (const short unsigned int *a, const short unsigned int *b,
 		  short unsigned int *c, LDPARMS * ldp);
-static int ecmp (_CONST short unsigned int *a, _CONST short unsigned int *b);
+static int ecmp (const short unsigned int *a, const short unsigned int *b);
 static int enormlz (short unsigned int *x);
 static int eshift (short unsigned int *x, int sc);
 static void eshup1 (register short unsigned int *x);
@@ -77,7 +77,7 @@ static void eshdn1 (register short unsigned int *x);
 static void eshdn8 (register short unsigned int *x);
 static void eshdn6 (register short unsigned int *x);
 static void eneg (short unsigned int *x);
-static void emov (register _CONST short unsigned int *a,
+static void emov (register const short unsigned int *a,
 		  register short unsigned int *b);
 static void eclear (register short unsigned int *x);
 static void einfin (register short unsigned int *x, register LDPARMS * ldp);
@@ -111,24 +111,24 @@ static void e113toe (short unsigned int *pe, short unsigned int *y,
 
 #if NE == 10
 /* 0.0 */
-static _CONST unsigned short ezero[NE] = { 0x0000, 0x0000, 0x0000, 0x0000,
+static const unsigned short ezero[NE] = { 0x0000, 0x0000, 0x0000, 0x0000,
   0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
 };
 
 /* 1.0E0 */
-static _CONST unsigned short eone[NE] = { 0x0000, 0x0000, 0x0000, 0x0000,
+static const unsigned short eone[NE] = { 0x0000, 0x0000, 0x0000, 0x0000,
   0x0000, 0x0000, 0x0000, 0x0000, 0x8000, 0x3fff,
 };
 
 #else
 
 /* 0.0 */
-static _CONST unsigned short ezero[NE] = {
+static const unsigned short ezero[NE] = {
   0, 0000000, 0000000, 0000000, 0000000, 0000000,
 };
 
 /* 1.0E0 */
-static _CONST unsigned short eone[NE] = {
+static const unsigned short eone[NE] = {
   0, 0000000, 0000000, 0000000, 0100000, 0x3fff,
 };
 
@@ -140,7 +140,7 @@ static _CONST unsigned short eone[NE] = {
  * messages is bound to the error codes defined
  * in mconf.h.
  */
-static _CONST char *_CONST ermsg[7] = {
+static const char *const ermsg[7] = {
   "unknown",			/* error code 0 */
   "domain",			/* error code 1 */
   "singularity",		/* et seq.      */
@@ -412,8 +412,10 @@ static void eaddm (short unsigned int *x, short unsigned int *y);
 static void esubm (short unsigned int *x, short unsigned int *y);
 static void emdnorm (short unsigned int *s, int lost, int subflg,
 		     long int exp, int rcntrl, LDPARMS * ldp);
+#if 0 /* Broken, unusable implementation of strtold */
 static int asctoeg (char *ss, short unsigned int *y, int oprec,
 		    LDPARMS * ldp);
+#endif
 static void enan (short unsigned int *nan, int size);
 #if LDBL_MANT_DIG == 24
 static void toe24 (short unsigned int *x, short unsigned int *y);
@@ -432,17 +434,17 @@ static int edivm (short unsigned int *den, short unsigned int *num,
 		  LDPARMS * ldp);
 static int emulm (short unsigned int *a, short unsigned int *b,
 		  LDPARMS * ldp);
-static int eisneg (_CONST short unsigned int *x);
-static int eisinf (_CONST short unsigned int *x);
-static void emovi (_CONST short unsigned int *a, short unsigned int *b);
+static int eisneg (const short unsigned int *x);
+static int eisinf (const short unsigned int *x);
+static void emovi (const short unsigned int *a, short unsigned int *b);
 static void emovo (short unsigned int *a, short unsigned int *b,
 		   LDPARMS * ldp);
 static void emovz (register short unsigned int *a,
 		   register short unsigned int *b);
 static void ecleaz (register short unsigned int *xi);
-static void eadd1 (_CONST short unsigned int *a, _CONST short unsigned int *b,
+static void eadd1 (const short unsigned int *a, const short unsigned int *b,
 		   short unsigned int *c, int subflg, LDPARMS * ldp);
-static int eisnan (_CONST short unsigned int *x);
+static int eisnan (const short unsigned int *x);
 static int eiisnan (short unsigned int *x);
 
 #ifdef DEC
@@ -473,7 +475,7 @@ eclear (register short unsigned int *x)
  */
 
 static void
-emov (register _CONST short unsigned int *a, register short unsigned int *b)
+emov (register const short unsigned int *a, register short unsigned int *b)
 {
   register int i;
 
@@ -506,7 +508,7 @@ eneg (short unsigned int *x)
  * else return zero.
  */
 static int
-eisneg (_CONST short unsigned int *x)
+eisneg (const short unsigned int *x)
 {
 
 #ifdef NANS
@@ -524,7 +526,7 @@ eisneg (_CONST short unsigned int *x)
  * else return zero.
  */
 static int
-eisinf (_CONST short unsigned int *x)
+eisinf (const short unsigned int *x)
 {
 
   if ((x[NE - 1] & 0x7fff) == 0x7fff)
@@ -542,7 +544,7 @@ eisinf (_CONST short unsigned int *x)
 /* Check if e-type number is not a number.
  */
 static int
-eisnan (_CONST short unsigned int *x)
+eisnan (const short unsigned int *x)
 {
 
 #ifdef NANS
@@ -612,9 +614,9 @@ einfin (register short unsigned int *x, register LDPARMS * ldp)
  * converting it to internal format.
  */
 static void
-emovi (_CONST short unsigned int *a, short unsigned int *b)
+emovi (const short unsigned int *a, short unsigned int *b)
 {
-  register _CONST unsigned short *p;
+  register const unsigned short *p;
   register unsigned short *q;
   int i;
 
@@ -1421,7 +1423,7 @@ mdfin:
 */
 
 static void
-esub (_CONST short unsigned int *a, _CONST short unsigned int *b,
+esub (const short unsigned int *a, const short unsigned int *b,
       short unsigned int *c, LDPARMS * ldp)
 {
 
@@ -1452,7 +1454,7 @@ esub (_CONST short unsigned int *a, _CONST short unsigned int *b,
 
 
 static void
-eadd1 (_CONST short unsigned int *a, _CONST short unsigned int *b,
+eadd1 (const short unsigned int *a, const short unsigned int *b,
        short unsigned int *c, int subflg, LDPARMS * ldp)
 {
   unsigned short ai[NI], bi[NI], ci[NI];
@@ -1563,7 +1565,7 @@ done:
 ;	ediv( a, b, c, ldp );	c = b / a
 */
 static void
-ediv (_CONST short unsigned int *a, _CONST short unsigned int *b,
+ediv (const short unsigned int *a, const short unsigned int *b,
       short unsigned int *c, LDPARMS * ldp)
 {
   unsigned short ai[NI], bi[NI];
@@ -1669,7 +1671,7 @@ dnzro2:
 ;	emul( a, b, c, ldp );	c = b * a
 */
 static void
-emul (_CONST short unsigned int *a, _CONST short unsigned int *b,
+emul (const short unsigned int *a, const short unsigned int *b,
       short unsigned int *c, LDPARMS * ldp)
 {
   unsigned short ai[NI], bi[NI];
@@ -1834,6 +1836,7 @@ e113toe (short unsigned int *pe, short unsigned int *y, LDPARMS * ldp)
 
 /* move out internal format to ieee long double */
 static void
+__attribute__ ((__unused__))
 toe113 (short unsigned int *a, short unsigned int *b)
 {
   register unsigned short *p, *q;
@@ -1970,6 +1973,7 @@ e64toe (short unsigned int *pe, short unsigned int *y, LDPARMS * ldp)
 
 /* move out internal format to ieee long double */
 static void
+__attribute__ ((__unused__))
 toe64 (short unsigned int *a, short unsigned int *b)
 {
   register unsigned short *p, *q;
@@ -2150,6 +2154,7 @@ etoe53 (x, e)
 }
 
 static void
+__attribute__ ((__unused__))
 toe53 (x, y)
      unsigned short *x, *y;
 {
@@ -2159,6 +2164,7 @@ toe53 (x, y)
 #else
 
 static void
+__attribute__ ((__unused__))
 toe53 (short unsigned int *x, short unsigned int *y)
 {
   unsigned short i;
@@ -2327,6 +2333,7 @@ e24toe (short unsigned int *pe, short unsigned int *y, LDPARMS * ldp)
 }
 
 static void
+__attribute__ ((__unused__))
 toe24 (short unsigned int *x, short unsigned int *y)
 {
   unsigned short i;
@@ -2415,7 +2422,7 @@ toe24 (short unsigned int *x, short unsigned int *y)
  *          -2 if either a or b is a NaN.
  */
 static int
-ecmp (_CONST short unsigned int *a, _CONST short unsigned int *b)
+ecmp (const short unsigned int *a, const short unsigned int *b)
 {
   unsigned short ai[NI], bi[NI];
   register unsigned short *p, *q;
@@ -2627,7 +2634,7 @@ normdn:
 #define MAXP 4096
 
 #if NE == 10
-static _CONST unsigned short etens[NTEN + 1][NE] = {
+static const unsigned short etens[NTEN + 1][NE] = {
   {0x6576, 0x4a92, 0x804a, 0x153f,
    0xc94c, 0x979a, 0x8a20, 0x5202, 0xc460, 0x7525,},	/* 10**4096 */
   {0x6a32, 0xce52, 0x329a, 0x28ce,
@@ -2656,7 +2663,7 @@ static _CONST unsigned short etens[NTEN + 1][NE] = {
    0x0000, 0x0000, 0x0000, 0x0000, 0xa000, 0x4002,},	/* 10**1 */
 };
 
-static _CONST unsigned short emtens[NTEN + 1][NE] = {
+static const unsigned short emtens[NTEN + 1][NE] = {
   {0x2030, 0xcffc, 0xa1c3, 0x8123,
    0x2de3, 0x9fde, 0xd2ce, 0x04c8, 0xa6dd, 0x0ad8,},	/* 10**-4096 */
   {0x8264, 0xd2cb, 0xf2ea, 0x12d4,
@@ -2685,7 +2692,7 @@ static _CONST unsigned short emtens[NTEN + 1][NE] = {
    0xcccc, 0xcccc, 0xcccc, 0xcccc, 0xcccc, 0x3ffb,},	/* 10**-1 */
 };
 #else
-static _CONST unsigned short etens[NTEN + 1][NE] = {
+static const unsigned short etens[NTEN + 1][NE] = {
   {0xc94c, 0x979a, 0x8a20, 0x5202, 0xc460, 0x7525,},	/* 10**4096 */
   {0xa74d, 0x5de4, 0xc53d, 0x3b5d, 0x9e8b, 0x5a92,},	/* 10**2048 */
   {0x650d, 0x0c17, 0x8175, 0x7586, 0xc976, 0x4d48,},
@@ -2701,7 +2708,7 @@ static _CONST unsigned short etens[NTEN + 1][NE] = {
   {0x0000, 0x0000, 0x0000, 0x0000, 0xa000, 0x4002,},	/* 10**1 */
 };
 
-static _CONST unsigned short emtens[NTEN + 1][NE] = {
+static const unsigned short emtens[NTEN + 1][NE] = {
   {0x2de4, 0x9fde, 0xd2ce, 0x04c8, 0xa6dd, 0x0ad8,},	/* 10**-4096 */
   {0x4925, 0x2de4, 0x3436, 0x534f, 0xceae, 0x256b,},	/* 10**-2048 */
   {0x87a6, 0xc0bd, 0xda57, 0x82a5, 0xa2a6, 0x32b5,},
@@ -2787,7 +2794,6 @@ _ldtoa_r (struct _reent *ptr, long double d, int mode, int ndigits,
   LDPARMS rnd;
   LDPARMS *ldp = &rnd;
   char *outstr;
-  char outbuf[NDEC + MAX_EXP_DIGITS + 10];
   union uconv du;
   du.d = d;
 
@@ -2834,6 +2840,8 @@ _ldtoa_r (struct _reent *ptr, long double d, int mode, int ndigits,
   if (ndigits > NDEC)
     ndigits = NDEC;
 
+  char outbuf[ndigits + MAX_EXP_DIGITS + 10];
+
   etoasc (e, outbuf, ndigits, mode, ldp);
   s = outbuf;
   if (eisinf (e) || eisnan (e))
@@ -2916,7 +2924,7 @@ stripspaces:
   for (_REENT_MP_RESULT_K (ptr) = 0;
        sizeof (_Bigint) - sizeof (__ULong) + j <= i; j <<= 1)
     _REENT_MP_RESULT_K (ptr)++;
-  _REENT_MP_RESULT (ptr) = Balloc (ptr, _REENT_MP_RESULT_K (ptr));
+  _REENT_MP_RESULT (ptr) = eBalloc (ptr, _REENT_MP_RESULT_K (ptr));
 
 /* Copy from internal temporary buffer to permanent buffer.  */
   outstr = (char *) _REENT_MP_RESULT (ptr);
@@ -2973,7 +2981,7 @@ etoasc (short unsigned int *x, char *string, int ndigits, int outformat,
 {
   long digit;
   unsigned short y[NI], t[NI], u[NI], w[NI];
-  _CONST unsigned short *p, *r, *ten;
+  const unsigned short *p, *r, *ten;
   unsigned short sign;
   int i, j, k, expon, rndsav, ndigs;
   char *s, *ss;
@@ -3104,6 +3112,8 @@ tnzro:
       else
 	{
 	  emovi (y, w);
+	  /* Note that this loop does not access the incoming string array,
+	   * which may be shorter than NDEC + 1 bytes! */
 	  for (i = 0; i < NDEC + 1; i++)
 	    {
 	      if ((w[NI - 1] & 0x7) != 0)
@@ -3328,7 +3338,7 @@ asctoeg (char *ss, short unsigned int *y, int oprec, LDPARMS * ldp)
   int k, trail, c, rndsav;
   long lexp;
   unsigned short nsign;
-  _CONST unsigned short *p;
+  const unsigned short *p;
   char *sp, *s, *lstr;
   int lenldstr;
   int mflag = 0;
@@ -3654,7 +3664,7 @@ aexit:
  *
  * efloor( x, y, ldp );
  */
-static _CONST unsigned short bmask[] = {
+static const unsigned short bmask[] = {
   0xffff,
   0xfffe,
   0xfffc,
@@ -3756,47 +3766,47 @@ eiremain (short unsigned int *den, short unsigned int *num, LDPARMS * ldp)
  */
 #ifdef MIEEE
 #if !defined(__mips)
-static _CONST unsigned short nan113[8] = {
+static const unsigned short nan113[8] = {
   0x7fff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff
 };
 
-static _CONST unsigned short nan64[6] = {
+static const unsigned short nan64[6] = {
   0x7fff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff
 };
-static _CONST unsigned short nan53[4] = { 0x7fff, 0xffff, 0xffff, 0xffff };
-static _CONST unsigned short nan24[2] = { 0x7fff, 0xffff };
+static const unsigned short nan53[4] = { 0x7fff, 0xffff, 0xffff, 0xffff };
+static const unsigned short nan24[2] = { 0x7fff, 0xffff };
 #elif defined(__mips_nan2008)	/* __mips */
-static _CONST unsigned short nan113[8] = { 0x7fff, 0x8000, 0, 0, 0, 0, 0, 0 };
-static _CONST unsigned short nan64[6] = { 0x7fff, 0xc000, 0, 0, 0, 0 };
-static _CONST unsigned short nan53[4] = { 0x7ff8, 0, 0, 0 };
-static _CONST unsigned short nan24[2] = { 0x7fc0, 0 };
+static const unsigned short nan113[8] = { 0x7fff, 0x8000, 0, 0, 0, 0, 0, 0 };
+static const unsigned short nan64[6] = { 0x7fff, 0xc000, 0, 0, 0, 0 };
+static const unsigned short nan53[4] = { 0x7ff8, 0, 0, 0 };
+static const unsigned short nan24[2] = { 0x7fc0, 0 };
 #else /* __mips && !__mips_nan2008 */
-static _CONST unsigned short nan113[8] = {
+static const unsigned short nan113[8] = {
   0x7fff, 0x7fff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff
 };
 
-static _CONST unsigned short nan64[6] = {
+static const unsigned short nan64[6] = {
   0x7fff, 0xbfff, 0xffff, 0xffff, 0xffff, 0xffff
 };
-static _CONST unsigned short nan53[4] = { 0x7ff7, 0xffff, 0xffff, 0xffff };
-static _CONST unsigned short nan24[2] = { 0x7fbf, 0xffff };
+static const unsigned short nan53[4] = { 0x7ff7, 0xffff, 0xffff, 0xffff };
+static const unsigned short nan24[2] = { 0x7fbf, 0xffff };
 #endif /* __mips && !__mips_nan2008 */
 #else /* !MIEEE */
 #if !defined(__mips) || defined(__mips_nan2008)
-static _CONST unsigned short nan113[8] = { 0, 0, 0, 0, 0, 0, 0x8000, 0x7fff };
-static _CONST unsigned short nan64[6] = { 0, 0, 0, 0, 0xc000, 0x7fff };
-static _CONST unsigned short nan53[4] = { 0, 0, 0, 0x7ff8 };
-static _CONST unsigned short nan24[2] = { 0, 0x7fc0 };
+static const unsigned short nan113[8] = { 0, 0, 0, 0, 0, 0, 0x8000, 0x7fff };
+static const unsigned short nan64[6] = { 0, 0, 0, 0, 0xc000, 0x7fff };
+static const unsigned short nan53[4] = { 0, 0, 0, 0x7ff8 };
+static const unsigned short nan24[2] = { 0, 0x7fc0 };
 #else /* __mips && !__mips_nan2008 */
-static _CONST unsigned short nan113[8] = {
+static const unsigned short nan113[8] = {
   0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0xffff, 0x7fff, 0x7fff
 };
 
-static _CONST unsigned short nan64[6] = {
+static const unsigned short nan64[6] = {
   0xffff, 0xffff, 0xffff, 0xffff, 0xbfff, 0x7fff
 };
-static _CONST unsigned short nan53[4] = { 0xffff, 0xffff, 0xffff, 0x7ff7 };
-static _CONST unsigned short nan24[2] = { 0xffff, 0x7fbf };
+static const unsigned short nan53[4] = { 0xffff, 0xffff, 0xffff, 0x7ff7 };
+static const unsigned short nan24[2] = { 0xffff, 0x7fbf };
 #endif /* __mips && !__mips_nan2008 */
 #endif /* !MIEEE */
 
@@ -3805,7 +3815,7 @@ static void
 enan (short unsigned int *nan, int size)
 {
   int i, n;
-  _CONST unsigned short *p;
+  const unsigned short *p;
 
   switch (size)
     {
