@@ -30,28 +30,10 @@ details. */
 #include <asm/socket.h>
 #include "cygwait.h"
 
-#define MAX_OVERLAPPED_WRITE_LEN (64 * 1024 * 1024)
-#define MIN_OVERLAPPED_WRITE_LEN (1 * 1024 * 1024)
-
 static const int CHUNK_SIZE = 1024; /* Used for crlf conversions */
 
 struct __cygwin_perfile *perfile_table;
 
-HANDLE NO_COPY fhandler_base_overlapped::asio_done;
-LONG NO_COPY fhandler_base_overlapped::asio_close_counter;
-
-void
-fhandler_base::reset (const fhandler_base *from)
-{
-  pc << from->pc;
-  ra.rabuf = NULL;
-  ra.ralen = 0;
-  ra.raixget = 0;
-  ra.raixput = 0;
-  ra.rabuflen = 0;
-  _refcnt = 0;
-}
-
 int
 fhandler_base::puts_readahead (const char *s, size_t len)
 {
@@ -196,8 +178,6 @@ fhandler_base::set_flags (int flags, int supplied_bin)
     bin = wbinary () || rbinary () ? O_BINARY : O_TEXT;
 
   openflags = flags | bin;
-  if (openflags & O_NONBLOCK)
-    was_nonblocking (true);
 
   bin &= O_BINARY;
   rbinary (bin ? true : false);
@@ -255,7 +235,7 @@ retry:
 		  break;
 		}
 	    }
-	  /*FALLTHRU*/
+	  fallthrough;
 	case STATUS_INVALID_DEVICE_REQUEST:
 	case STATUS_INVALID_PARAMETER:
 	case STATUS_INVALID_HANDLE:
@@ -265,6 +245,7 @@ retry:
 	      len = (size_t) -1;
 	      break;
 	    }
+	  fallthrough;
 	default:
 	  __seterrno_from_nt_status (status);
 	  len = (size_t) -1;
@@ -466,7 +447,7 @@ fhandler_base::open_with_arch (int flags, mode_t mode)
     {
       if (!archetype->get_handle ())
 	{
-	  copyto (archetype);
+	  archetype->copy_from (this);
 	  archetype_usecount (1);
 	  archetype->archetype = NULL;
 	  usecount = 0;
@@ -483,14 +464,15 @@ fhandler_base::open_with_arch (int flags, mode_t mode)
 	      strcpy (name, get_name ());
 	    }
 	  fhandler_base *arch = archetype;
-	  archetype->copyto (this);
+	  copy_from (archetype);
 	  if (name)
 	    set_name (name);
 	  archetype = arch;
 	  archetype_usecount (1);
 	  usecount = 0;
 	}
-      open_setup (flags);
+      if (!open_setup (flags))
+	api_fatal ("open_setup failed, %E");
     }
 
   close_on_exec (flags & O_CLOEXEC);
@@ -579,10 +561,12 @@ fhandler_base::open (int flags, mode_t mode)
       access = READ_CONTROL | FILE_READ_ATTRIBUTES;
       break;
     case query_write_control:
-      access = READ_CONTROL | WRITE_OWNER | WRITE_DAC | FILE_WRITE_ATTRIBUTES;
+      access = READ_CONTROL | WRITE_OWNER | WRITE_DAC
+	       | (pc.fs_is_samba () ? 0 : FILE_WRITE_ATTRIBUTES);
       break;
     case query_write_dac:
-      access = READ_CONTROL | WRITE_DAC | FILE_WRITE_ATTRIBUTES;
+      access = READ_CONTROL | WRITE_DAC
+	       | (pc.fs_is_samba () ? 0 : FILE_WRITE_ATTRIBUTES);
       break;
     case query_write_attributes:
       access = READ_CONTROL | FILE_WRITE_ATTRIBUTES;
@@ -619,13 +603,20 @@ fhandler_base::open (int flags, mode_t mode)
   else
     create_disposition = (flags & O_CREAT) ? FILE_OPEN_IF : FILE_OPEN;
 
-  if (get_device () == FH_FS)
+  if (get_device () == FH_FS
+#ifdef __WITH_AF_UNIX
+      || get_device () == FH_UNIX
+#endif
+      )
     {
-      /* Add the reparse point flag to known repares points, otherwise we
+      /* Add the reparse point flag to known reparse points, otherwise we
 	 open the target, not the reparse point.  This would break lstat. */
       if (pc.is_known_reparse_point ())
 	options |= FILE_OPEN_REPARSE_POINT;
+    }
 
+  if (get_device () == FH_FS)
+    {
       /* O_TMPFILE files are created with delete-on-close semantics, as well
 	 as with FILE_ATTRIBUTE_TEMPORARY.  The latter speeds up file access,
 	 because the OS tries to keep the file in memory as much as possible.
@@ -801,9 +792,10 @@ fhandler_base::fd_reopen (int, mode_t)
   return NULL;
 }
 
-void
+bool
 fhandler_base::open_setup (int)
 {
+  return true;
 }
 
 /* states:
@@ -1226,88 +1218,6 @@ fhandler_base::close ()
   return res;
 }
 
-DWORD WINAPI
-flush_async_io (void *arg)
-{
-  fhandler_base_overlapped *fh = (fhandler_base_overlapped *) arg;
-  debug_only_printf ("waiting for write I/O for %s", fh->get_name ());
-  DWORD nbytes;
-  bool res = GetOverlappedResult (fh->get_output_handle (),
-				  fh->get_overlapped (), &nbytes, true);
-  debug_printf ("finished waiting for I/O from %s, res %d", fh->get_name (),
-		res);
-  fh->close ();
-  delete fh;
-
-  InterlockedDecrement (&fhandler_base_overlapped::asio_close_counter);
-  SetEvent (fhandler_base_overlapped::asio_done);
-
-  _my_tls._ctinfo->auto_release ();
-  return 0;
-}
-
-void
-fhandler_base_overlapped::flush_all_async_io ()
-{
-  while (asio_close_counter > 0)
-    if (WaitForSingleObject (asio_done, INFINITE) != WAIT_OBJECT_0)
-      {
-	system_printf ("WaitForSingleObject failed, possible data loss in pipe, %E");
-	break;
-      }
-  asio_close_counter = 0;
-  if (asio_done)
-    CloseHandle (asio_done);
-}
-
-/* Start a thread to handle closing of overlapped asynchronous I/O since
-   Windows amazingly does not seem to always flush I/O on close.  */
-void
-fhandler_base_overlapped::check_later ()
-{
-  set_close_on_exec (true);
-  char buf[MAX_PATH];
-  if (!asio_done
-      && !(asio_done = CreateEvent (&sec_none_nih, false, false,
-				    shared_name (buf, "asio",
-						 GetCurrentProcessId ()))))
-    api_fatal ("CreateEvent failed, %E");
-
-  InterlockedIncrement (&asio_close_counter);
-  if (!new cygthread(flush_async_io, this, "flasio"))
-    api_fatal ("couldn't create a thread to track async I/O, %E");
-  debug_printf ("started thread to handle asynchronous closing for %s", get_name ());
-}
-
-int
-fhandler_base_overlapped::close ()
-{
-  int res;
-  int writer = (get_access () & GENERIC_WRITE);
-  /* Need to treat non-blocking I/O specially because Windows appears to
-     be brain-dead.  We're checking here if the descriptor was ever set
-     to nonblocking, rather than checking if it's nonblocking at close time.
-     The reason is that applications may switch back to blocking (for the
-     sake of some other application accessing this descriptor) without
-     performaing any further I/O.  These applications would suffer data
-     loss, which this workaround is trying to fix. */
-  if (writer && was_nonblocking () && has_ongoing_io ())
-    {
-      clone (HEAP_3_FHANDLER)->check_later ();
-      res = 0;
-    }
-  else
-    {
-     /* Cancelling seems to be necessary for cases where a reader is
-	 still executing when a signal handler performs a close.  */
-      if (!writer)
-	CancelIo (get_handle ());
-      destroy_overlapped ();
-      res = fhandler_base::close ();
-    }
-  return res;
-}
-
 int
 fhandler_base::ioctl (unsigned int cmd, void *buf)
 {
@@ -1402,7 +1312,7 @@ fhandler_base::init (HANDLE f, DWORD a, mode_t bin)
 }
 
 int
-fhandler_base::dup (fhandler_base *child, int)
+fhandler_base::dup (fhandler_base *child, int flags)
 {
   debug_printf ("in fhandler_base dup");
 
@@ -1411,7 +1321,7 @@ fhandler_base::dup (fhandler_base *child, int)
     {
       if (!DuplicateHandle (GetCurrentProcess (), get_handle (),
 			    GetCurrentProcess (), &nh,
-			    0, TRUE, DUPLICATE_SAME_ACCESS))
+			    0, !(flags & O_CLOEXEC), DUPLICATE_SAME_ACCESS))
 	{
 	  debug_printf ("dup(%s) failed, handle %p, %E",
 			get_name (), get_handle ());
@@ -1425,14 +1335,6 @@ fhandler_base::dup (fhandler_base *child, int)
   return 0;
 }
 
-int
-fhandler_base_overlapped::dup (fhandler_base *child, int flags)
-{
-  int res = fhandler_base::dup (child, flags) ||
-	    ((fhandler_base_overlapped *) child)->setup_overlapped ();
-  return res;
-}
-
 int fhandler_base::fcntl (int cmd, intptr_t arg)
 {
   int res;
@@ -1566,6 +1468,7 @@ fhandler_base::fhandler_base () :
   _refcnt (0),
   openflags (0),
   unique_id (0),
+  select_sem (NULL),
   archetype (NULL),
   usecount (0)
 {
@@ -1652,13 +1555,6 @@ fhandler_base::fixup_after_fork (HANDLE parent)
     del_my_locks (after_fork);
 }
 
-void
-fhandler_base_overlapped::fixup_after_fork (HANDLE parent)
-{
-  setup_overlapped ();
-  fhandler_base::fixup_after_fork (parent);
-}
-
 void
 fhandler_base::fixup_after_exec ()
 {
@@ -1667,12 +1563,6 @@ fhandler_base::fixup_after_exec ()
     del_my_locks (after_exec);
   mandatory_locking (false);
 }
-void
-fhandler_base_overlapped::fixup_after_exec ()
-{
-  setup_overlapped ();
-  fhandler_base::fixup_after_exec ();
-}
 
 bool
 fhandler_base::is_nonblocking ()
@@ -1686,8 +1576,6 @@ fhandler_base::set_nonblocking (int yes)
   int current = openflags & O_NONBLOCK_MASK;
   int new_flags = yes ? (!current ? O_NONBLOCK : current) : 0;
   openflags = (openflags & ~O_NONBLOCK_MASK) | new_flags;
-  if (new_flags)
-    was_nonblocking (true);
 }
 
 int
@@ -1951,272 +1839,32 @@ fhandler_base::fpathconf (int v)
   return -1;
 }
 
-/* Overlapped I/O */
-
-int __reg1
-fhandler_base_overlapped::setup_overlapped ()
+NTSTATUS
+fhandler_base::npfs_handle (HANDLE &nph)
 {
-  OVERLAPPED *ov = get_overlapped_buffer ();
-  memset (ov, 0, sizeof (*ov));
-  set_overlapped (ov);
-  ov->hEvent = CreateEvent (&sec_none_nih, true, true, NULL);
-  io_pending = false;
-  return ov->hEvent ? 0 : -1;
-}
+  static NO_COPY SRWLOCK npfs_lock;
+  static NO_COPY HANDLE npfs_dirh;
 
-void __reg1
-fhandler_base_overlapped::destroy_overlapped ()
-{
-  OVERLAPPED *ov = get_overlapped ();
-  if (ov && ov->hEvent)
-    {
-      SetEvent (ov->hEvent);
-      CloseHandle (ov->hEvent);
-      ov->hEvent = NULL;
-    }
-  io_pending = false;
-  get_overlapped () = NULL;
-}
-
-bool __reg1
-fhandler_base_overlapped::has_ongoing_io ()
-{
-  if (!io_pending)
-    return false;
-
-  if (!IsEventSignalled (get_overlapped ()->hEvent))
-    return true;
-  io_pending = false;
-  DWORD nbytes;
-  GetOverlappedResult (get_output_handle (), get_overlapped (), &nbytes, false);
-  return false;
-}
-
-fhandler_base_overlapped::wait_return __reg3
-fhandler_base_overlapped::wait_overlapped (bool inres, bool writing, DWORD *bytes, bool nonblocking, DWORD len)
-{
-  if (!get_overlapped ())
-    return inres ? overlapped_success : overlapped_error;
-
-  wait_return res = overlapped_unknown;
-  DWORD err;
-  if (inres)
-    /* handle below */;
-  else if ((err = GetLastError ()) != ERROR_IO_PENDING)
-    res = overlapped_error;
-  else if (!nonblocking)
-    /* handle below */;
-  else if (!writing)
-    SetEvent (get_overlapped ()->hEvent);	/* Force immediate WFMO return */
-  else
-    {
-      *bytes = len;				/* Assume that this worked */
-      io_pending = true;			/*  but don't allow subsequent */
-      res = overlapped_success;			/*  writes until completed */
-    }
-  if (res == overlapped_unknown)
-    {
-      DWORD wfres = cygwait (get_overlapped ()->hEvent);
-      HANDLE h = writing ? get_output_handle () : get_handle ();
-      BOOL wores;
-      if (isclosed ())
-	{
-	  switch (wfres)
-	    {
-	    case WAIT_OBJECT_0:
-	      err = ERROR_INVALID_HANDLE;
-	      break;
-	    case WAIT_SIGNALED:
-	      err = ERROR_INVALID_AT_INTERRUPT_TIME;
-	      break;
-	    default:
-	      err = GetLastError ();
-	      break;
-	    }
-	  res = overlapped_error;
-	}
-      else
-	{
-	  /* Cancelling here to prevent races.  It's possible that the I/O has
-	     completed already, in which case this is a no-op.  Otherwise,
-	     WFMO returned because 1) This is a non-blocking call, 2) a signal
-	     arrived, or 3) the operation was cancelled.  These cases may be
-	     overridden by the return of GetOverlappedResult which could detect
-	     that I/O completion occurred.  */
-	  CancelIo (h);
-	  wores = GetOverlappedResult (h, get_overlapped (), bytes, false);
-	  err = GetLastError ();
-	  ResetEvent (get_overlapped ()->hEvent);	/* Probably not needed but CYA */
-	  debug_printf ("wfres %u, wores %d, bytes %u", wfres, wores, *bytes);
-	  if (wores)
-	    res = overlapped_success;	/* operation succeeded */
-	  else if (wfres == WAIT_OBJECT_0 + 1)
-	    {
-	      err = ERROR_INVALID_AT_INTERRUPT_TIME; /* forces an EINTR below */
-	      debug_printf ("signal");
-	      res = overlapped_error;
-	    }
-	  else if (wfres == WAIT_OBJECT_0 + 2)
-	    pthread::static_cancel_self ();		/* never returns */
-	  else if (nonblocking)
-	    res = overlapped_nonblocking_no_data;	/* more handling below */
-	  else
-	    {
-	      debug_printf ("GetOverlappedResult failed, h %p, bytes %u, %E", h, *bytes);
-	      res = overlapped_error;
-	    }
-	}
-    }
-
-  if (res == overlapped_success)
-    {
-      debug_printf ("normal %s, %u bytes ispipe() %d", writing ? "write" : "read", *bytes, ispipe ());
-      if (*bytes == 0 && !writing && ispipe ())
-	res = overlapped_nullread;
-    }
-  else if (res == overlapped_nonblocking_no_data)
-    {
-      *bytes = (DWORD) -1;
-      set_errno (EAGAIN);
-      debug_printf ("no data to read for nonblocking I/O");
-    }
-  else if (err == ERROR_HANDLE_EOF || err == ERROR_BROKEN_PIPE)
-    {
-      debug_printf ("EOF, %E");
-      *bytes = 0;
-      res = overlapped_success;
-      if (writing && err == ERROR_BROKEN_PIPE)
-	raise (SIGPIPE);
-    }
-  else
-    {
-      debug_printf ("res %u, Win32 Error %u", (unsigned) res, err);
-      *bytes = (DWORD) -1;
-      __seterrno_from_win_error (err);
-      if (writing && err == ERROR_NO_DATA)
-	raise (SIGPIPE);
-    }
-
-  return res;
-}
-
-void __reg3
-fhandler_base_overlapped::raw_read (void *ptr, size_t& len)
-{
-  DWORD nbytes;
-  bool keep_looping;
-  do
-    {
-      bool res = ReadFile (get_handle (), ptr, len, &nbytes,
-			   get_overlapped ());
-      switch (wait_overlapped (res, false, &nbytes, is_nonblocking ()))
-	{
-	case overlapped_nullread:
-	  keep_looping = true;
-	  break;
-	default:	/* Added to quiet gcc */
-	case overlapped_success:
-	case overlapped_error:
-	  keep_looping = false;
-	  break;
-	}
-    }
-  while (keep_looping);
-  len = (nbytes == (DWORD) -1) ? (size_t) -1 : (size_t) nbytes;
-}
+  NTSTATUS status = STATUS_SUCCESS;
+  OBJECT_ATTRIBUTES attr;
+  IO_STATUS_BLOCK io;
 
-ssize_t __reg3
-fhandler_base_overlapped::raw_write (const void *ptr, size_t len)
-{
-  size_t nbytes;
-  if (has_ongoing_io ())
+  /* Lockless after first call. */
+  if (npfs_dirh)
     {
-      set_errno (EAGAIN);
-      nbytes = (size_t) -1;
+      nph = npfs_dirh;
+      return STATUS_SUCCESS;
     }
-  else
+  AcquireSRWLockExclusive (&npfs_lock);
+  if (!npfs_dirh)
     {
-      size_t chunk;
-      if (!max_atomic_write || len < max_atomic_write)
-	chunk = MIN (len, INT_MAX);
-      else if (is_nonblocking ())
-	chunk = len = max_atomic_write;
-      else
-	chunk = max_atomic_write;
-
-      /* MSDN "WriteFile" contains the following note: "Accessing the output
-         buffer while a write operation is using the buffer may lead to
-	 corruption of the data written from that buffer.  [...]  This can
-	 be particularly problematic when using an asynchronous file handle.
-	 (https://msdn.microsoft.com/en-us/library/windows/desktop/aa365747)
-
-	 MSDN "Synchronous and Asynchronous I/O" contains the following note:
-	 "Do not deallocate or modify [...] the data buffer until all
-	 asynchronous I/O operations to the file object have been completed."
-	 (https://msdn.microsoft.com/en-us/library/windows/desktop/aa365683)
-
-	 This problem is a non-issue for blocking I/O, but it can lead to
-	 problems when using nonblocking I/O.  Consider:
-	 - The application uses a static buffer in repeated calls to
-	   non-blocking write.
-	 - The previous write returned with success, but the overlapped I/O
-	   operation is ongoing.
-	 - The application copies the next set of data to the static buffer,
-	   thus overwriting data still accessed by the previous write call.
-	 --> potential data corruption.
-
-	 What we do here is to allocate a per-fhandler buffer big enough
-	 to perform the maximum atomic operation from, copy the user space
-	 data over to this buffer and then call NtWriteFile on this buffer.
-	 This decouples the write operation from the user buffer and the
-	 user buffer can be reused without data corruption issues.
-
-	 Since no further write can occur while we're still having ongoing
-	 I/O, this should be reasanably safe.
-
-	 Note: We only have proof that this problem actually occurs on Wine
-	 yet.  However, the MSDN language indicates that this may be a real
-	 problem on real Windows as well. */
-      if (is_nonblocking ())
-	{
-	  if (!atomic_write_buf)
-	    atomic_write_buf = cmalloc_abort (HEAP_BUF, max_atomic_write);
-	  ptr = memcpy (atomic_write_buf, ptr, chunk);
-	}
-
-      nbytes = 0;
-      DWORD nbytes_now = 0;
-      /* Write to fd in smaller chunks, accumulating a total.
-	 If there's an error, just return the accumulated total
-	 unless the first write fails, in which case return value
-	 from wait_overlapped(). */
-      while (nbytes < len)
-	{
-	  size_t left = len - nbytes;
-	  size_t len1;
-	  if (left > chunk)
-	    len1 = chunk;
-	  else
-	    len1 = left;
-	  bool res = WriteFile (get_output_handle (), ptr, len1, &nbytes_now,
-				get_overlapped ());
-	  switch (wait_overlapped (res, true, &nbytes_now,
-				   is_nonblocking (), len1))
-	    {
-	    case overlapped_success:
-	      ptr = ((char *) ptr) + chunk;
-	      nbytes += nbytes_now;
-	      break;
-	    case overlapped_error:
-	      len = 0;		/* terminate loop */
-	    case overlapped_unknown:
-	    case overlapped_nullread:
-	    case overlapped_nonblocking_no_data:
-	      break;
-	    }
-	}
-      if (!nbytes)
-	nbytes = (nbytes_now == (DWORD) -1) ? (size_t) -1 : nbytes_now;
+      InitializeObjectAttributes (&attr, &ro_u_npfs, 0, NULL, NULL);
+      status = NtOpenFile (&npfs_dirh, FILE_READ_ATTRIBUTES | SYNCHRONIZE,
+			   &attr, &io, FILE_SHARE_READ | FILE_SHARE_WRITE,
+			   0);
     }
-  return nbytes;
+  ReleaseSRWLockExclusive (&npfs_lock);
+  if (NT_SUCCESS (status))
+    nph = npfs_dirh;
+  return status;
 }
